#!/usr/bin/perl -w
#
# popDiversity.pl v1.0
# Last modified October 4, 2016 11:05
# (c) Olaf R.P. Bininda-Emonds
#
# Usage: popDiversity.pl -f<filename> or -l<filename> -g<filename>
#		 [-i<f|g|n|p|s>] [-n<name>] [-p<path>] [-u] [-v] [-h]
#   options: -f<filename> = file containing raw sequence information
#			 -l<filename> = text logfile from PAUP* listing 1) full taxon status
#							for data set and 2) pairwise distances between all
#							sequences
#			 -g<filename> = tab-delimited text file of form sequence name<tab>
#							group membership; different group memberships can be
#							placed in subsequent columns with each column being
#							analyzed individually
#			 -d<distance> = PAUP* distance model chosen from MEAN, ABS,	
#							P (default), JC, F81, TAJNEI, K2P, F84, HKY85, K3P,
#							TAMNEI, or GTR
#			 -n<name> = base name to apply to result files (default = popDiv for
#						-l and filename for -f)
#			 -p<path> = create appropriate log file using paup (default = paup64
#						in $PATH)
#			 -u = also process undefined distances (marked with a *)
#            -v = verbose output
#            -h = print this message and quit

use strict;
use Math::Trig;
use File::Basename;

# Set user-input defaults and associated parameters
	# I/O variables
		my $matrixFile;
		my $groupFile;
		my $seqFile;
			my ($dataSource, $dir, $ext);
		my $baseName = "popDiv";
		my $userDistance = "P";
			my @paupDistances = qw(MEAN ABS P JC F81 TAJNEI K2P F84 HKY85 K3P TAMNEI GTR);
 			my %paupDistances;
 				$paupDistances{$_} = 1 foreach (@paupDistances);			

	# pairwiseReader variables
		my $printForm = "column";	# Options are column and matrix
		my $diagPrint = "below";	# Options are above, below, and both (latter for matrix only)
	
		my ($aboveFlag, $belowFlag) = (0, 0);
		my ($rowTaxon, $columnTaxon) = (0, 0);
		my ($startColumn, $endColumn);
		my (%aboveDiag, %belowDiag);
		
		my %taxonName;
		my $numTaxa = 0;
		my $ignoreUndefined = 1;
		my $paupVersion = "~/bin/paup64";
		
	# Group definitions
		my (%groupPresent, %groupNames, %groupID);
			my %namePresent;
		my $maxLevel = 0;
		my %distances;
		
	# seqRead variables
		my $inputType = "";	# Options are "fasta", "GenBank", "nexus", "phylip", and "Se-Al"
 		my $seqType = "DNA";
		my $globalGenCode = 0;
			my $globalOverride = 0;

		my (@accNum, %nameLabel, %sequence, %geneticCode, %accPresent);
		my (@charsetList, %charsetStart, %charsetEnd);
		my (%deletedSeq, %finalSeq);
		my $seqCount;
		my $ntax;
		my $skipDuplicates = 1;
		my (%specimenCount, @rootNames);
 		my $gapChange = 1;

	# Translation variables and genetic codes
		my %transTable = ('1' => 'standard',
						  '2' => 'vertebrate mitochondrial',
						  '3' => 'yeast mitochondrial',
						  '4' => 'mold, protozoan and colenterate mitochondrial and mycoplasam/spiroplasma',
						  '5' => 'invertebrate mitochondrial',
						  '6' => 'ciliate, dasycladacean and hexamita nuclear',
						  '9' => 'echinoderm mitochondrial',
						  '10' => 'euplotid nuclear',
						  '11' => 'bacterial and plant plastid',
						  '12' => 'alternative yeast nuclear',
						  '13' => 'ascidian mitochondrial',
						  '14' => 'alternative flatworm mitochondrial',
						  '15' => 'Blepharisma nuclear',
						  '16' => 'chlorophycean mitochondrial',
						  '21' => 'trematode mitochondrial',
						  '22' => 'Scenedesmus obliquus mitochondrial',
						  '23' => 'Thraustochytrium mitochondrial');
		my %gb2seal = ('1' => '0',
					   '2' => '1',
					   '3' => '2',
					   '4' => '3',
					   '4' => '4',
					   '5' => '5',
					   '6' => '6',
					   '9' => '7',
					   '10' => '8',
					   '11' => '9',
					   '12' => '10',
					   '13' => '11',
					   '14' => '12',
					   '15' => '13',
					   '16' => '1',
					   '21' => '1',
					   '22' => '1',
					   '23' => '1');
		my %seal2gb = ('0' => '1',
					   '1' => '2',
					   '2' => '3',
					   '3' => '4',
					   '4' => '4',
					   '5' => '5',
					   '6' => '6',
					   '7' => '9',
					   '8' => '10',
					   '9' => '11',
					   '10' => '12',
					   '11' => '13',
					   '12' => '14',
					   '13' => '15');

	# Miscellaneous variables
		my @errorList;
		my $verbose = 0;
		my $quiet = 1;
		my $debug = 0;
		my $perlScript = "popDiversity.pl";
			my $callString = join(' ', $perlScript, @ARGV);
		my $version = "1.0";
		my $usage = "Usage: $perlScript -f<filename> or -l<filename> -g<filename>\n";
			$usage .= "       [-i<f|g|n|p|s>] [-n<name>] [-p<path>] [-u] [-v] [-h]";
		my $options .= "Options: -f<filename> = file containing raw sequence information\n";
			$options .= "         -l<filename> = text logfile from PAUP* listing 1) full taxon status\n";
			$options .= "                        for data set and 2) pairwise distances between all\n";
			$options .= "                        sequences\n";
			$options .= "         -g<filename> = tab-delimited text file of form sequence name<tab>\n";
			$options .= "                        group membership; different group memberships can be\n";
			$options .= "                        placed in subsequent columns with each column being\n";
			$options .= "                        analyzed individually\n";
			$options .= "         -d<distance> = PAUP* distance model chosen from MEAN, ABS,\n";
			$options .= "                        P (default), JC, F81, TAJNEI, K2P, F84, HKY85, K3P,\n";
			$options .= "         -n<name> = base name to apply to result files (default = popDiv for\n";
			$options .= "                    -l and filename for -f)\n";
			$options .= "         -p<path> = create appropriate log file using paup\n";
			$options .= "                     (default = $paupVersion in \$PATH)\n";
			$options .= "         -u = also process undefined distances (marked with a *)\n";
			$options .= "         -h = print this message and quit\n";
			$options .= "         -v = verbose output";

# Process switches
	for (my $i = 0; $i <= $#ARGV; $i++)
		{
		if ($ARGV[$i] =~ /^-f(.+)/)
			{
			$seqFile = $1;
			push @errorList, "ERROR: Sequence file $seqFile does not exist.\n" unless (-e $seqFile);
			}
		elsif ($ARGV[$i] =~ /^-l(.+)/)
			{
			$matrixFile = $1;
			push @errorList, "ERROR: PAUP* logfile $matrixFile does not exist.\n" unless (-e $matrixFile);
			}
		elsif ($ARGV[$i] eq "-if")
			{
			$inputType = "fasta";
			}
		elsif ($ARGV[$i] eq "-in")
			{
			$inputType = "nexus";
			}
		elsif ($ARGV[$i] eq "-ip")
			{
			$inputType = "phylip";
			}
		elsif ($ARGV[$i] eq "-is")
			{
			$inputType = "Se-Al";
			}
		elsif ($ARGV[$i] =~ /^-g(.+)/)
			{
			$groupFile = $1;
			push @errorList, "ERROR: Tab-delimited text file with group definitions $groupFile does not exist.\n" unless (-e $groupFile);
			}
		elsif ($ARGV[$i] =~ /^-n(.+)/)
			{
			$baseName = $1;
			}
		elsif ($ARGV[$i] =~ /^-d(\w+)/)
			{
			$userDistance = $1;
			push @errorList, "ERROR: Distance model $userDistance not recognized.\n" unless (defined $paupDistances{$userDistance});
			}
		elsif ($ARGV[$i] =~ /^-p(.+)/)
			{
			$paupVersion = $1;
			}
		elsif ($ARGV[$i] eq "-u")
			{
			$ignoreUndefined = 0;
			}
		elsif ($ARGV[$i] eq "-v")
			{
			$verbose = 1;
			}
		elsif ($ARGV[$i] eq "-x")
			{
			$debug = $verbose = 1;
			}
		elsif ($ARGV[$i] eq "-h")
			{
			print "$usage\n";
			print "Version: $version\n";
			print "$options\n";
			exit(0);
			}
		else
			{
			print "Don't understand argument: $ARGV[$i]\n";
			print "$usage\n";
			print "Version: $version\n";
			exit(1); 
			}
		}
		
	# Error checking
		push @errorList, "ERROR: Must specify tab-delimited text file containing group identities\n" unless (defined $groupFile);
		
		if (defined $seqFile)	# Check for working version of paup
			{
			my ($paup1, $paup2);
			system ("$paupVersion --v > paupCheck.txt 2>&1");
				open (PAUP, "<paupCheck.txt") or die "Cannot open file paupCheck.txt ($!)\n";
					while (<PAUP>)
						{
						chomp;
						next unless ($_);
						if ($_ =~ /4\.0(\w)(\d+)/)
							{
							$paup1 = $1;
							$paup2 = $2;
							}
						}
				close PAUP;
				push @errorList, "ERROR: PAUP* not found at $paupVersion\n." if (not defined $paup2);
				push @errorList, "ERROR: Version of PAUP* (4.0$paup1$paup2) not recent enough (requires >4.0b10).\n" if ($paup2  <= 10);
				unlink <paupCheck.txt> unless ($debug);
			}
		else
			{
			push @errorList, "ERROR: Must specify PAUP* log file containing pairwise distances\n" unless (defined $matrixFile);
			}
	
		if (@errorList)
			{
			print "$_" foreach (@errorList);
			exit(1);
			}

		print "Script was called as (" . localtime() . "):\n\t$callString\n" if (defined $callString);
			
# Read in group assignments
	setLineBreak($groupFile);
	open (GROUP, "<$groupFile") or die "Cannot open $groupFile to read group assignments ($!).\n";
		while (<GROUP>)
			{
			chomp;
			next unless ($_);
			
			my @groupLine = split(/\t/, $_);
			
			my $seqName = shift(@groupLine);
				if (defined $seqFile)
					{
					$seqName =~ s/\W+/_/g;
					$seqName =~ s/\_+/_/g;
					}
				$namePresent{$seqName} = 1;
				for (my $level = 0; $level < scalar(@groupLine); $level++)
					{
					last if (not defined $groupLine[$level]);
					$maxLevel = $level if ($level > $maxLevel);
					
					$groupPresent{$groupLine[$level]}{$level}++;
						push @ { $groupNames{$level} }, $groupLine[$level] if ($groupPresent{$groupLine[$level]}{$level} == 1);

					$groupID{$seqName}{$level} = $groupLine[$level];
					}
			}
	close GROUP;

# Read in (or create) and process matrix
	if ($matrixFile)
		{
		matrixRead($matrixFile);
		}
	else
		{
		($dataSource, $dir, $ext) = fileparse($seqFile, qr/\.\D.*/);
			my $baseFile = $dataSource;

		setLineBreak($seqFile);
		seqRead($seqFile);
			die "\tERROR: Could not read in sequences from file $seqFile\n" if (not @accNum);
		
			# Ensure that sequence has full length, check for group assignments, and correct names
				my $maxLength = 0;
				foreach my $entry (@accNum)
					{
					next if ($deletedSeq{$entry});
					$maxLength = length($sequence{$entry}) if (length($sequence{$entry}) > $maxLength);
					}

				my @missingGroup;
				foreach my $entry (@accNum)
					{
					next if ($deletedSeq{$entry});

					$nameLabel{$entry} =~ s/\W+/_/g;
						$nameLabel{$entry} =~ s/\_+/_/g;
					$sequence{$entry} .= "-" x ($maxLength - length($sequence{$entry}));
					push @missingGroup, $nameLabel{$entry} if (not defined $namePresent{$nameLabel{$entry}});
					}
					
				if (@missingGroup)
					{
					print "\nERROR: The following sequence(s) are missing group assignments:\n";
						print "\t$_\n" foreach (@missingGroup);
					exit(1);
					}

		# Print data file in simple nexus format with necessary instructions
			$ntax = scalar(@accNum);
			
			my $nexusOut = "popPaup.nex";
			open (NEX, ">$nexusOut") or die "Cannot write data to nexus file $nexusOut ($!)\n";
				print NEX "#NEXUS\n\n";
				print NEX "[File created from $seqFile using $perlScript v$version on ".localtime()."]\n\n";
				print NEX "begin data;\n";
					print NEX "\tdimensions ntax = $ntax nchar = $maxLength;\n";
					print NEX "\tformat datatype = $seqType gap = - missing = ?;\n";
					print NEX "\tmatrix\n\n";
						foreach my $entry (@accNum)
							{
							next if ($deletedSeq{$entry});
							print NEX "$nameLabel{$entry}\t$sequence{$entry}\n";
						}
				print NEX "\t;\nend;\n";
			
				print NEX "\nbegin paup;\n";
#					print NEX "\tlog file = popOut_log.txt replace;\n";
					print NEX "\ttstatus full = yes;\n";
					print NEX "\n\tdset distance = $userDistance;\n";
						print NEX "\t\tshowdist;\n";
				print NEX "end;\n";
				print NEX "\nquit;\n";
			close NEX;

			if (-e "/Developer/Tools/SetFile")
				{
				system ("/Developer/Tools/SetFile -t 'TEXT' $nexusOut");
				system ("/Developer/Tools/SetFile -c 'PAUP' $nexusOut");
				}
				
			system ("$paupVersion $nexusOut > $baseFile\_log.txt");
				unlink "$nexusOut" unless ($debug);

		matrixRead("$baseFile\_log.txt");
#			unlink "popOut_log.txt" unless ($debug);
		}

	&matrixPrint;
	
# Print within and among group distances for each level
	my $outFile = "popDivResults_$baseName\.txt";
	open (OUT, ">$outFile") or die "Cannot write results to $outFile ($!)\n";

	for (my $level = 0; $level <= $maxLevel; $level++)
		{
		my $distFile = "pairwiseDistances_$baseName\.$level\.txt";
		open (DIST, ">$distFile") or die "Cannot write pairwise distances for level $level to $distFile ($!)\n";
		
		my @compList = @ { $groupNames{$level} };
			my $numComps = scalar(@compList)**2 - scalar(@compList);
			print OUT "\nComparing groups for level $level (n = $numComps comparisons) ...\n";
			
		for (my $i = 0; $i < scalar(@ { $groupNames{$level} }) - 1; $i++)
			{
			my $group1 = $groupNames{$level}[$i];
				print OUT "\tPopulation $group1\n";
			my ($average1, $variance1, $median1, $mode1) = statsPack(@{ $distances{$level}{$group1}});
			my @distances1 = sort {$b <=> $a} @{ $distances{$level}{$group1}};
				push @distances1, 0 if (not @distances1);	# If population only represented by a single specimen
			
			for (my $j = $i + 1; $j < scalar(@ { $groupNames{$level} }); $j++)
				{
				my $group2 = $groupNames{$level}[$j];
					print OUT "\t\tVersus population $group2\n";
				my ($average2, $variance2, $median2, $mode2) = statsPack(@{ $distances{$level}{$group2}});
				my @distances2 = sort {$b <=> $a} @{ $distances{$level}{$group2}};
					push @distances2, 0 if (not @distances2);	# If population only represented by a single specimen
				
				my @compList;
					push @compList, $group1, $group2;
						@compList = sort @compList;
				my $compType = join(' - ', @compList);

				print OUT "\t\t\tAmong population stats:\n";
					my ($average, $variance, $median, $mode) = statsPack(@{ $distances{$level}{$compType}});
					my @distancesAmong = sort {$a <=> $b} @{ $distances{$level}{$compType}};
						printf OUT "\t\t\t\tPairwise comparisons: %s\n", scalar(@{ $distances{$level}{$compType}});
							next unless (@{ $distances{$level}{$compType}});	# No pairwise comparisons
						printf OUT "\t\t\t\tAverage: %.5f\n", $average;
						print OUT "\t\t\t\tVariance: ";
						if ($variance)
							{
							printf OUT ($variance < 0.00001) ? "%.5e\n" : "%.5f\n", $variance;
							}
						else
							{
							print OUT "n/a\n";
							}
						printf OUT "\t\t\t\tMedian: %.5f\n", $median;
						printf OUT "\t\t\t\tMode: %s\n", $mode;
					print DIST "$compType\t$_\n" foreach (@{ $distances{$level}{$compType}});
					
				print OUT "\t\t\tBarcoding stats:\n";
					my $barCoding1 = ($average1) ? sprintf("%.1f", $average / $average1) : "infinite";
						print OUT "\t\t\t\tBarcode gap for $group1: $barCoding1\n";
					my $barCoding2 = ($average2) ? sprintf("%.1f", $average / $average2) : "infinite";
						print OUT "\t\t\t\tBarcode gap for $group2: $barCoding2\n";
						
					print OUT "\n\t\t\t\tMax-min overlap for $group1";
						print OUT " not" unless ($distancesAmong[0] > $distances1[0]) ;
						print OUT " fulfilled ($distances1[0] vs $distancesAmong[0])\n";
					print OUT "\t\t\t\tMax-min overlap for $group2";
						print OUT " not" unless ($distancesAmong[0] > $distances2[0]) ;
						print OUT " fulfilled ($distances2[0] vs $distancesAmong[0])\n";
						
					my ($t1, $df1, $P1) = unpairedT(\@{ $distances{$level}{$compType}}, \@{ $distances{$level}{$group1}});
						if ($t1 eq "undefined")
							{
							print OUT "\n\t\t\t\tUnpaired t for $group1 is undefined\n";
							}
						else
							{
							printf OUT "\n\t\t\t\tUnpaired t for $group1 %.3f (df = $df1; uncorrected P = %.4f", $t1, $P1;
							print OUT " *" if ($P1 < 0.05 / $numComps);
							print OUT ")\n";
							}
					my ($t2, $df2, $P2) = unpairedT(\@{ $distances{$level}{$compType}}, \@{ $distances{$level}{$group2}});
						if ($t2 eq "undefined")
							{
							print OUT "\n\t\t\t\tUnpaired t for $group1 is undefined\n";
							}
						else
							{
							printf OUT "\t\t\t\tUnpaired t for $group2 %.3f (df = $df2; uncorrected P = %.4f", $t2, $P2;
							print OUT " *" if ($P2 < 0.05 / $numComps);
							print OUT ")\n";
							}
				}
			}

	print OUT "\n\tWithin population stats:\n";
		foreach my $compType (@compList)
			{
			print OUT "\t\t$compType:\n";
			
			my ($average, $variance, $median, $mode) = statsPack(@{ $distances{$level}{$compType}});
				printf OUT "\t\t\tPairwise comparisons: %s ($groupPresent{$compType}{$level} sequences)\n", scalar(@{ $distances{$level}{$compType}});
				if (scalar(@{ $distances{$level}{$compType}} > 1))
					{
					printf OUT "\t\t\tAverage: %.5f\n", $average;
					printf OUT ($variance < 0.00001) ? "\t\t\tVariance: %.5e\n" : "\t\t\tVariance: %.5f\n", $variance;
					printf OUT "\t\t\tMedian: %.5f\n", $median;
					printf OUT "\t\t\tMode: %s\n", $mode;
					}

			print DIST "$compType\t$_\n" foreach (@{ $distances{$level}{$compType}});
			}
		close DIST;
		}		
	close OUT;	

exit(0);

# Subroutines used in script

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}

sub matrixRead
	{
	my $inFile = shift;

	my $prevLine = "";
	my ($taxonRead, $matrixRead) = (0, 0);
		my ($tstatusFlag, $matrixFlag) = (0 ,0);
	my %taxonRawName;
	
	# Variables to guard against PAUP* line wrapping
		my $wrapFlag = 0;
		my $numColumns;
		my $lastTaxon = 0;
		my $matrixLine;
	
	setLineBreak($inFile);
	open (IN, "<$inFile") or die "Cannot open $inFile to read matrix ($!).\n";
		while (<IN>)
			{
			chomp;
			print "$_\n" if ($debug);
			
			$paupVersion = $1 if ($_ =~ /version 4.0(\w+)/i);
			
			# Set reading flags
				if ($_ =~ /distance matrix$/ or $_ =~ /^No data for comparing taxa/)
					{
					$taxonRead = 0;
					$matrixRead = 1;
					$matrixFlag = 1;
					}

				if ($_ =~ /Current status of all taxa/)
					{
					$taxonRead = 1;
					$matrixRead = 0;
					$tstatusFlag = 1;
					print "Reading in taxon status ...\n" if ($debug);
					}
					
				next if ($_ =~ /^No data for comparing taxa/);
					
			# Read in taxon IDs; accounts for spaces in taxon names
				if ($taxonRead)
					{
					my $nameLength = 16;	# PAUP* only prints out first 16 or 30 characters in distance matrix
						$nameLength = 30 if ($paupVersion =~/a/);
						
					next unless ($_ =~ /\s?\d+/);
						$_ =~ s/^\s+//;
					
					my @taxonLine = split(/\s+/, $_);
						my $taxonID = shift(@taxonLine);
						pop(@taxonLine);	# Get rid of last entry indicating taxon status
						$taxonRawName{$taxonID} = join(' ', @taxonLine);
							$taxonRawName{$taxonID} = substr($taxonRawName{$taxonID}, 0, $nameLength);	# PAUP* only prints out first 16 or 30 characters in distance matrix
						$taxonName{$taxonID} = join('_', @taxonLine);
						print "\t$taxonID = $taxonName{$taxonID} ($taxonRawName{$taxonID})\n" if ($debug);
					}

			# Read in matrix
				if ($matrixRead)
					{
					die "ERROR: Failed to read in taxon status information (PAUP command = tstatus full).\n" unless ($tstatusFlag);
					# Skip lines not related to matrix
						$_ =~ s/^\s+//;
#							$_ =~ s/\*//g;
							next if ($_ =~ /^[a-zA-Z]/);
							next if ($_ =~ /^\*/);
							next if ($_ =~ /characters are included/);

 					# Mark blank lines, which indicate new portion of matrix coming			
						if (not $_)
							{
							$prevLine = "blank";
							next;
							}
							
					# Read in matrix
						my $readLine = $_;
							my @readLine = split(/\s+/, $readLine);
						
						if ($prevLine eq "blank")	# Read in header line
							{
							$startColumn = shift(@readLine);
							$endColumn = pop(@readLine);
								$endColumn = $startColumn if (not defined $endColumn);	# Only one column in matrix
								$numTaxa = $endColumn;
							$numColumns = $endColumn - $startColumn + 1;
							$prevLine = "";
							next;
							}
						else	# Process matrix
							{
							if ($debug)
								{
								print "Read: $readLine\n";
								print "\ttaxonID: $readLine[0]" if (defined $readLine[0]);
								print "\ttaxonName: $taxonName{$readLine[0]}" if (defined $readLine[1]);
								print "\ttaxonRawName: $taxonRawName{$readLine[0]}" if (defined $taxonRawName{$readLine[0]});
								print "\n\n";
								}
							# Safety to check if PAUP* has wrapped part of matrix on to a new line
								if (defined $readLine[0] and defined $taxonRawName{$readLine[0]} and $taxonRawName{$readLine[0]} =~ /^$readLine[1]/)	# Check that start of line formats as taxonID<white space>taxonName
									{
									$rowTaxon = $readLine[0];
										$readLine =~ s/$taxonRawName{$rowTaxon}/$taxonName{$rowTaxon}/;	# Correct line for taxon names with spaces
									$matrixLine = $readLine;
										print "\tNormal line: $readLine ($startColumn)\n" if ($debug);
									}
								else	# Wrapped line; append to previous line and reparse
									{
									$matrixLine .= " $readLine";
									print "\t\tMatrix: $matrixLine ($startColumn)\n" if ($debug);
									}

							@readLine = split(/\s+/, $matrixLine);

							my $columnTaxon = $startColumn - 1;
							for (my $column = 2; $column < scalar(@readLine); $column++)
								{
								$columnTaxon++;
								my $value = $readLine[$column];
									$columnTaxon = $rowTaxon if ($value eq "-");
									print "$rowTaxon\t$columnTaxon\t$value\n" if ($debug);
#print "$rowTaxon\t$columnTaxon\t$value\n";
#print "$taxonName{$rowTaxon}\t$taxonName{$columnTaxon}\t$value\n";
									if ($value =~ /\*$/)
										{
										next if ($ignoreUndefined);
										$value =~ s/\*$//;
										}
								
								if ($columnTaxon > $rowTaxon)
									{
									$aboveDiag{$rowTaxon}{$columnTaxon} = $aboveDiag{$columnTaxon}{$rowTaxon} = $value;
									$aboveDiag{$taxonName{$rowTaxon}} {$taxonName{$columnTaxon}} = $aboveDiag{$taxonName{$columnTaxon}} {$taxonName{$rowTaxon}} = $value;
									$aboveFlag = 1;
									}
								if ($columnTaxon < $rowTaxon)
									{
									$belowDiag{$rowTaxon}{$columnTaxon} = $value;
									$belowDiag{$columnTaxon}{$rowTaxon} = $value;
									$belowDiag{$taxonName{$rowTaxon}} {$taxonName{$columnTaxon}} = $belowDiag{$taxonName{$columnTaxon}} {$taxonName{$rowTaxon}} = $value;
									$belowFlag = 1;
									}
								}
							}
					}
			}
	# Error checking
		die "ERROR: Failed to read in pairwise distance matrix.\n" unless ($matrixFlag);

	return;
	}

sub matrixPrint	# Note: altered from basic version
	{
	# Do error processing
		die "ERROR: Above diagonal entries not defined in input.\n" if (not $aboveFlag and $diagPrint eq "above");
		die "ERROR: Below diagonal entries not defined in input.\n" if (not $belowFlag and $diagPrint eq "below");

	# Print out header
		if ($printForm eq "matrix")
			{
			for (my $column = 1; $column <= $numTaxa; $column++)
				{
				print "\t$column";
				}
				print "\n";
			}
		if ($printForm eq "column")
			{
#			print "Taxon 1\tTaxon 2\tValue\n";
			}
	
	for (my $row = 1; $row <= $numTaxa; $row++)
		{
		$taxonName{$row} = $row if (not defined $taxonName{$row});	# Safety if no names defined in file

		for (my $column = 1; $column <= $numTaxa; $column++)
			{
			next if (not defined $aboveDiag{$row}{$column} and $printForm eq "column" and $diagPrint eq "above");
			next if (not defined $belowDiag{$row}{$column} and $printForm eq "column" and $diagPrint eq "below");

			$taxonName{$column} = $column if (not defined $taxonName{$column});	# Safety if no names defined in file
#			print "$taxonName{$row}" if ($printForm eq "column" or ($printForm eq "matrix" and $column == 1));
#			print "\t";
#			print "$taxonName{$column}\t" if ($printForm eq "column");
			
			for (my $level = 0; $level <= $maxLevel; $level++)	# Note: altered from basic version
				{
				next unless (defined $groupID{$taxonName{$row}}{$level} and defined $groupID{$taxonName{$column}}{$level});
				my $compType = $groupID{$taxonName{$row}}{$level};
#					$compType = "Among groups"
					if ($groupID{$taxonName{$row}}{$level} ne $groupID{$taxonName{$column}}{$level})
						{
						my @compList;
							push @compList, $groupID{$taxonName{$row}}{$level}, $groupID{$taxonName{$column}}{$level};
							@compList = sort @compList;
							$compType = join(' - ', @compList);
						}

				if ($row > $column)
					{
					push @ { $distances{$level}{$compType}}, $belowDiag{$row}{$column} if (defined $belowDiag{$row}{$column} and ($diagPrint eq "below" or $diagPrint eq "both"));
#					print "$belowDiag{$row}{$column}" if (defined $belowDiag{$row}{$column} and ($diagPrint eq "below" or $diagPrint eq "both"));
					}
				elsif ($row < $column)
					{
					push @ { $distances{$level}{$compType}}, $aboveDiag{$row}{$column} if (defined $aboveDiag{$row}{$column} and ($diagPrint eq "above" or $diagPrint eq "both"));
#					print "$aboveDiag{$row}{$column}" if (defined $aboveDiag{$row}{$column} and ($diagPrint eq "above" or $diagPrint eq "both"));
					}
				else
					{
#					print "-";
					}
				}
#			print "\n" if ($printForm eq "column");
			}
		print "\n" unless ($printForm eq "column");
		}
	}

sub statsPack
	{
	my @rawData = sort {$a <=> $b} @_;
		my $numObs = scalar(@rawData);
			print "numObs: $numObs\n" if ($debug);
		return unless (@rawData);
	
	my ($sum, $sumSquares);
	my %count;
	
	foreach my $dataPoint (@rawData)
		{
		$sum += $dataPoint;
		$sumSquares += $dataPoint * $dataPoint;
		
		$count{$dataPoint}++;
		}
		
	my $average = $sum / scalar(@rawData);
		print "average: $average\n" if ($debug);
	my $median;
		if (int($numObs / 2) * 2 == $numObs)
			{
			$median = ($rawData[($numObs / 2)] + $rawData[($numObs / 2) - 1]) / 2;
			}
		else
			{
			$median = $rawData[int($numObs / 2)];
			}
		print "median: $median\n" if ($debug);
	my $variance = "";
		if ($numObs > 1)
			{
			$variance = ($sumSquares - $sum * $average) / ($numObs - 1);
			$variance = 0 if ($variance < 0);
			}
		print "variance: $variance\n" if ($debug);
	my $mode;
		my @modeList = sort { $count{$b} <=> $count{$a} } keys %count;
		my $modeCount = $count{$modeList[0]};
			if ($modeCount == 1)
				{
				$mode = join(',', @modeList);
				}
			else
				{
				$mode = $modeList[0];
				foreach (my $i = 1; $i < scalar(@modeList); $i++)
					{
					if ($count{$modeList[$i]} == $modeCount)
						{
						$mode .= "," . $modeList[$i];
						}
					else
						{
						last;
						}
					}
				}
		print "mode: $mode\n" if ($debug);
		
	return $average, $variance, $median, $mode;
	}

sub unpairedT	# Assumes equal variances
	{
	my ($array1, $array2) = @_;
		my @array1 = @$array1;
		my @array2 = @$array2;
		
	my ($average1, $variance1, $median1, $mode1) = statsPack(@$array1);
		my $count1 = scalar(@$array1);

	my ($average2, $variance2, $median2, $mode2) = statsPack(@$array2);
		my $count2 = scalar(@$array2);
		
	if (not $variance1 or not $variance2 or ($variance1 == 0 and $variance2 == 0))
		{
		return "undefined", "undefined", "undefined";
		}

	my $t = ($average1 - $average2);
	my $df = $count1 + $count2 - 2;
	
	if ($count1 == $count2)	# Equal sample sizes and equal variances
		{
		eval { $t /= sqrt(0.5 * ($variance1 + $variance2)) };
			$t /= 0.001 if ($@);
		$t /= sqrt(2 / ($count1 + $count2));
		}
	else	# Unequal sample sizes
		{
#		if ($variance1 != $variance2)	# Unequal variances; should be tested using BrownForsythe test or Levene's tests
#			{
#			$t /= sqrt($variance1 / $count1 + $variance2 / $count2);
#			$df = ($variance1 / $count1 + $variance2 / $count2)**2;
#			$df /= (($variance1 / $count1)**2 / ($count1 - 1)) + (($variance2 / $count2)**2 / ($count2 - 1));
#			}
#		else
#			{
			eval { $t /= sqrt( (((($count1 - 1) * $variance1) + (($count2 - 1) * $variance2))) / ($count1 + $count2 - 2) ) };
				$t /= 0.001 if ($@);
			$t /= sqrt(1/$count1 + 1/$count2);
#			}
		}
		
	my $lowerP = _subtprob($df, $t);
#		$lowerP = 1 - $lowerP if ($t > 0);	# Actually not necessary given that stats not calculated in this instance
		
	return ($t, $df, $lowerP);
	}

sub _subtprob	# Calculates upper probability from Student's t-distribution; taken from the CPAN Perl module Statistics::Distributions by Michael Kospach
	{
	my ($n, $x) = @_;

	my ($a,$b);
	my $w = atan2($x / sqrt($n), 1);
	my $z = cos($w) ** 2;
	my $y = 1;

	for (my $i = $n-2; $i >= 2; $i -= 2)
		{
		$y = 1 + ($i-1) / $i * $z * $y;
		} 

	if ($n % 2 == 0)
		{
		$a = sin($w)/2;
		$b = .5;
		}
	else
		{
		$a = ($n == 1) ? 0 : sin($w)*cos($w)/pi;
		$b= .5 + $w/pi;
		}
	return max(0, 1 - $b - $a * $y);
	}
	
sub max	# Finds maximum value in an array; taken from the CPAN Perl module Statistics::Distributions by Michael Kospach
	{
	my $max = shift;
	my $next;
	while (@_)
		{
		$next = shift;
		$max = $next if ($next > $max);
		}	
	return $max;
	}

sub seqRead
	{
	my $seqFile = shift;
	undef %sequence;

	print "\nReading in sequence data from file $seqFile (type is $inputType) ...\n" if ($inputType and not $quiet);
	setLineBreak($seqFile);
	open (SEQ, "<$seqFile") or die "Cannot open file containing sequences, $seqFile ($!)\n";
		my ($header, $tempAcc, $tempName, $tempSeq);
		my $fastaAcc;
		my (%nexusSpecies, %nexusAcc, $nexusRead, $commentFlag);
		my ($phylipLineCount, $phylipTaxa, $phylipChars, %phylipSeq);
		my $sealCode;
		my ($sealDelFlag, $owner) = (0, 0);
		my ($gbAcc, $gbRead);
		my $macBlock = 0;
		my %accCount;

		while (<SEQ>)
			{
			chomp;
			my $lineRead = $_;
			next unless ($lineRead);
			
			# Autodetect sequence format
				if (not $inputType)
					{
					$inputType = "fasta" if ($lineRead =~ /^>/);
					$inputType = "nexus" if ($lineRead =~ /\#nexus/i);
					$inputType = "phylip" if ($lineRead =~ /^\s*\d+\s+\d+/);
					$inputType = "Se-Al" if ($lineRead =~ /^\s*Database=\{/i);
					$inputType = "GenBank" if ($lineRead =~ /^\s*LOCUS/);
					print "\nReading in sequence data from file $seqFile (type determined to be $inputType) ...\n" if ($inputType);
					}
			
			if ($inputType eq "nexus")
				{
				# Check if charset statement present
					if ($lineRead =~ /charset/i)
						{
						$lineRead =~ s/\s+//g;
						$lineRead =~ s/;$//;

						my ($charsetName, $charsetBounds) = split('=', $lineRead);
							$charsetName =~ s/charset//i;
								push @charsetList, $charsetName;
							if ($charsetBounds =~ /(\d+)-*(\d*)/)
								{
								$charsetStart{$charsetName} = $1;
								if ($2)
									{
									$charsetEnd{$charsetName} = $2;
									}
								else
									{
									$charsetEnd{$charsetName} = $1;
									}
								}
						}
						
				# Otherwise block out MacClade / PAUP blocks
					$macBlock = 1 if ($lineRead =~ /begin macclade;/i or $lineRead =~ /begin paup;/i);
					$macBlock = 0 if ($macBlock and $lineRead =~ /end;/i);
					next if ($macBlock);
						
				# Otherwise only read in data lines or charset statements
					if ($lineRead =~ /^\s*matrix/i)
						{
						$nexusRead = 1;
						next;
						}
					$nexusRead = 0 if ($lineRead =~ /;\s*$/);
					next unless ($nexusRead);
					
					# Remove MacClade sequence lengths
						$lineRead =~ s/\[\d+\]$//;

					$commentFlag = 1 if ($lineRead =~ /\[/);
					if ($lineRead =~ /\]/)
						{
						$commentFlag = 0;
						next;
						}
					next if ($commentFlag);

					next unless ($lineRead =~ /a/i or $lineRead =~ /c/i or $lineRead =~ /g/i or $lineRead =~ /t/i or $lineRead =~ /n/i or $lineRead =~ /\?/ or $lineRead =~ /\-/);

				# Clean up input line
					$lineRead =~ s/^\s+//;
#					$lineRead =~ s/\'//g;

				my @nexusLine = split(/\s+/, $lineRead);
					my $species = shift(@nexusLine);
					my $tempAcc;
						if ($species =~ /^'/)	# Contains non-alphanumerics, possibly including spaces, possibly including accession numbers
							{
							until ($species =~ /'$/)
								{
								$species .= "_" . shift(@nexusLine);
								}
							# Check for accession numbers
								if ($species =~ /(\(?\w+\d+\.?\d*\)?)/)
									{
									$tempAcc = $1;
										$tempAcc =~ s/\(//;
										$tempAcc =~ s/\)//;
									$species =~ s/$tempAcc//;
										$species =~ s/\(\)//;
										$species =~ s/^'//;
										$species =~ s/'$//;
									}
							}

						$species =~ s/\s+/_/g;
						$species =~ s/\_+/_/g;
						$species =~ s/\_$//g;
						
						$nexusAcc{$species} = $tempAcc if (defined $tempAcc);

					my $seq = join('', @nexusLine);
						$seq =~ s/\s+//g;
						$seqType = "protein" if ($seq =~ /E/i or $seq =~ /Q/i or $seq =~ /I/i or $seq =~ /L/i or $seq =~ /F/i or $seq =~ /P/i);

				if (not defined $nexusSpecies{$species})
					{
					$nexusSpecies{$species} = 1;
					$seqCount++;
					$nexusAcc{$species} = "tAlign_".$seqCount unless (defined $nexusAcc{$species});
					push @accNum, $nexusAcc{$species};
						$nameLabel{$nexusAcc{$species}} = $species;
						$sequence{$nexusAcc{$species}} = uc($seq);
						$geneticCode{$nexusAcc{$species}} = $globalGenCode;
					}
				else	# Sequences are in interleaved format; append sequence
					{
					$sequence{$nexusAcc{$species}} .= uc($seq);
					}
				}

			if ($inputType eq "fasta")
				{
				if ($lineRead =~/^\s*>/)
					{
					my $species;
					$seqCount++;
					(my $tempSpecies = $lineRead) =~ s/^\s*>//;
					
					# Parse different formats
						if ($tempSpecies =~ /^Mit\.\s+/)	# Entry comes from European RNA project
							{
							$tempSpecies =~ s/^Mit\.\s+//i;	# To fix entries from European RNA project
							my @speciesInfo = split(/\s+/, $tempSpecies);
								$species = join('_', $speciesInfo[0], $speciesInfo[1]);
							if (defined $speciesInfo[2])
								{
								$fastaAcc = $speciesInfo[2];
								$accCount{$fastaAcc}++;
									if ($accCount{$fastaAcc} > 1)
										{
										print "\nWARNING: Accession number $fastaAcc used more than once";
											print "; skipping this and subsequent entries" if ($skipDuplicates);
										print "\n";
										}
								}
							else
								{
								$fastaAcc = "tAlign_".$seqCount;
								$accCount{$fastaAcc}++;
								}
							}
						elsif ($tempSpecies =~ /^(\w+\.\d+\.\d+)/)	# Raw SILVA / ARB RNA project
							{
							$fastaAcc = $1;
							$accCount{$fastaAcc}++;
								if ($accCount{$fastaAcc} > 1)
									{
									print "\nWARNING: Accession number $fastaAcc used more than once";
										print "; skipping this and subsequent entries" if ($skipDuplicates);
									print "\n";
									}
									
							($species = $tempSpecies) =~ s/^\w+\.\d+\.\d+//;
								$species =~ s/^\s+//;
								$species =~ s/\s+/_/g;
							}
						else
							{
							# Parse Seaview specific code
								if ($tempSpecies =~ /\s+\/transl_table=(\d+)$/)
									{
									$sealCode = $1;
									$tempSpecies =~ s/\s+\/transl_table=\d+$//;
									}
								if ($tempSpecies =~ /_(\(?[A-Z]{1,2}_*\d{5,6}\.?\d*\)?)$/)
									{
									my $tempAcc = $1;
										$tempSpecies =~ s/$tempAcc//;
										$tempSpecies =~ s/\(\)//;
										$tempSpecies =~ s/_$//;
										$tempSpecies .= "\t$tempAcc";
									}
							my @speciesLine = split(/\s+/, $tempSpecies);
							if (($speciesLine[$#speciesLine] =~ /^\(?[A-Z]{1,2}_*\d{5,6}\.?\d*\)?$/ or $speciesLine[$#speciesLine] =~ /\w+\.\d+\.\d+/) and scalar(@speciesLine) > 1)	# Check whether last entry is an accession number
								{
								$fastaAcc = pop (@speciesLine);
								$fastaAcc =~ s/^\(//g;
								$fastaAcc =~ s/\)$//g;
								$accCount{$fastaAcc}++;
									if ($accCount{$fastaAcc} > 1)
										{
										print "\nWARNING: Accession number $fastaAcc used more than once";
											if ($skipDuplicates)
												{
												print "; skipping this and subsequent entries\n";
												}
											else
												{
												print "; assigning temporary accession number\n";
												$fastaAcc = "tAlign_" . $seqCount;
												}
										}
								}
							else
								{
								$fastaAcc = "tAlign_".$seqCount;
								$accCount{$fastaAcc}++;
								}
							$species = join('_', @speciesLine);
								$species = "Sequence_".$seqCount if ($species eq "");
							}
					push @accNum, $fastaAcc unless ($accCount{$fastaAcc} > 1 and $skipDuplicates);
						$geneticCode{$fastaAcc} = $globalGenCode;
					($nameLabel{$fastaAcc} = $species) =~ s/\s+/_/;
						$nameLabel{$fastaAcc} =~ s/\_+/_/;
					}
				else
					{
					next if ($accCount{$fastaAcc} > 1 and $skipDuplicates);
					$lineRead =~ s/\s+//g;
						$seqType = "protein" if ($lineRead =~ /E/i or $lineRead =~ /Q/i or $lineRead =~ /I/i or $lineRead =~ /L/i or $lineRead =~ /F/i or $lineRead =~ /P/i);
					$sequence{$fastaAcc} .= uc($lineRead);
					}
				$geneticCode{$fastaAcc} = $sealCode;
				}

			if ($inputType eq "Se-Al")
				{
				my $header;
				$sealDelFlag = 1 if ($lineRead =~ /MCoL/ or $lineRead =~ /^\}\;/);	# Se-Al sometimes places deleted species at end of file; do not read in remainder of file
					next if ($sealDelFlag == 1);
				next unless ($lineRead =~ /NumSites=/ or $lineRead =~ /Owner=/ or $lineRead =~ /Name=/ or $lineRead =~ /Accession=/ or $lineRead =~ /Sequence=/ or $lineRead =~/GeneticCode=/ or $lineRead =~ /Frame=/);
					$lineRead =~ s/^\s+//;

				if ($lineRead =~ /^Owner\s*\=\s*(\d+)/)
					{
					$owner = $1;
					}
				if ($lineRead =~ /^Accession=/ and $owner == 2)
					{
					$seqCount++;
					if ($lineRead =~ /null/ or $lineRead =~ /\"\"/)
						{
						$tempAcc = "tAlign_" . $seqCount;
						$accCount{$tempAcc}++;
						}
					else
						{
						($header, $tempAcc) = split (/=/, $lineRead);
							$tempAcc =~ s/\"//g;
							$tempAcc =~ s/;//g;
							$accCount{$tempAcc}++;
								if ($accCount{$tempAcc} > 1)
									{
									print "\nWARNING: Accession number $tempAcc used more than once";
										if ($skipDuplicates)
											{
											print "; skipping this and subsequent entries\n";
											}
										else
											{
											print "; assigning temporary accession number\n";
											$tempAcc = "tAlign_" . $seqCount;
											}
									print "\n";
									}
						}
					}
				if ($lineRead =~ /^Name=/ and $owner == 2)
					{
					($header, $tempName) = split (/=/, $lineRead);
						$tempName =~ s/\"//g;
						$tempName =~ s/\s*;//g;
						$tempName =~ s/\s+/_/g;
						$tempName =~ s/\_+/_/g;
					}
				if ($lineRead =~ /^GeneticCode=(\d+)/ and $owner == 2)
					{
					$sealCode = $1;
					$geneticCode{$tempAcc} = $seal2gb{$sealCode};
					}
				if ($lineRead =~ /^Sequence=/ and $owner == 2)
					{
					next if ($accCount{$tempAcc} > 1 and $skipDuplicates);
					($header, $tempSeq) = split (/=/, $lineRead);
						$tempSeq =~ s/\"//g;
						$tempSeq =~ s/;//g;
						$tempSeq =~ s/\s+//g;
						next if ($tempSeq eq "null");
						
					push @accNum, $tempAcc unless ($accCount{$tempAcc} > 1 and $skipDuplicates);
					$nameLabel{$tempAcc} = $tempName;
					$sequence{$tempAcc} = uc($tempSeq);
						$seqType = "protein" if ($tempSeq =~ /E/i or $tempSeq =~ /Q/i or $tempSeq =~ /I/i or $tempSeq =~ /L/i or $tempSeq =~ /F/i or $tempSeq =~ /P/i);
					}
				if ($lineRead =~ /^Frame=(\d)/)	# Correct for reading frame
					{
					my $readingFrame = $1;
					$sequence{$tempAcc} = "--" . $sequence{$tempAcc} if ($readingFrame == 2);
					$sequence{$tempAcc} = "-" . $sequence{$tempAcc} if ($readingFrame == 3);
					}
				}

			if ($inputType eq "phylip")
				{
				if ($lineRead =~ /^\s*(\d+)\s+(\d+)/)
					{
					$seqCount = $phylipTaxa = $1;
					$phylipChars = $2;
					$phylipLineCount = 0;
					}
				else
					{
					$phylipLineCount++;
					
					$lineRead =~ s/\s//g;
					
					$phylipSeq{$phylipLineCount} .= $lineRead;
					
					$phylipLineCount = 0 if ($phylipLineCount == $phylipTaxa);
					}
				}

			if ($inputType eq "GenBank")
				{
				# Get species name and accession number
					# Pure GenBank format
						$gbAcc = $1 if ($lineRead =~ /^\s*ACCESSION\s+(\w+\d+)/);
						if ($lineRead =~ /^\s*ORGANISM\s+/)
							{
							$seqCount++;
								$gbAcc = "tAlign_" . $seqCount if (not defined $gbAcc);
							$lineRead =~ s/^\s+//;
							my @orgLine = split(/\s+/, $lineRead);
								my $header = shift (@orgLine);
								$nameLabel{$gbAcc} = join('_', @orgLine);
								$accCount{$gbAcc}++;
							}
					# BioEdit format
						if ($lineRead =~ /^\s*TITLE/ and not defined ($gbAcc))
							{
							$seqCount++;
								$gbAcc = "tAlign_" . $seqCount;
							my @accLine = split (/\s+/, $lineRead);
								$gbAcc = $1 if ($accLine[2] =~ /^\((\w+\d+)\)/);
							$nameLabel{$gbAcc} = $accLine[1];
							$accCount{$gbAcc}++;
							}

				if ($lineRead =~ /^\s*ORIGIN/)
					{
					if ($accCount{$gbAcc} > 1)
						{
						print "\nWARNING: Accession number $gbAcc used more than once";
						if ($skipDuplicates)
							{
							print "; skipping this and subsequent entries\n";
							next;
							}
						else
							{
							print "; assigning temporary accession number\n";
							$gbAcc = "tAlign_" . $seqCount;
							$gbRead = 1;
							}
						}
					else
						{
						$gbRead = 1;
						$seqCount++;
						}
					next;
					}

				if ($lineRead =~ /^\s*\/\//)	# End of accession; process
					{
					$gbRead = 0;
					next unless ($gbAcc);
					
					push @accNum, $gbAcc unless ($accCount{$gbAcc} > 1);
					$geneticCode{$gbAcc} = $globalGenCode;
					$nameLabel{$gbAcc} =~ s/\s+/_/g;
						$nameLabel{$gbAcc} =~ s/\_+/_/g;
					$sequence{$gbAcc} =~ s/\d//g;
						$sequence{$gbAcc} =~ s/\s+//g;
						$sequence{$gbAcc} =~ s/\~/-/g if ($gapChange);
					undef $gbAcc;
					}
					
				next unless ($gbRead);
				
				if ($gbRead and $lineRead =~ /^\s+\d+/)
					{
					$sequence{$gbAcc} .= uc($lineRead);
					}
				}
			}
	close SEQ;
	
	if ($inputType eq "phylip")	# Postprocess input to derive taxon names and sequence; accounts for both sequential and extended formatting
		{
		for (my $i = 1; $i <= $phylipTaxa; $i++)
			{
			my $phylipAcc = "tAlign_" . $i;
			
			push @accNum, $phylipAcc;
			$geneticCode{$phylipAcc} = $globalGenCode;
			
			# Derive taxon name and sequence
				$sequence{$phylipAcc} = uc(substr($phylipSeq{$i}, 0 - $phylipChars));
					$seqType = "protein" if ($sequence{$phylipAcc} =~ /E/i or $sequence{$phylipAcc} =~ /Q/i or $sequence{$phylipAcc} =~ /I/i or $sequence{$phylipAcc} =~ /L/i or $sequence{$phylipAcc} =~ /F/i or $sequence{$phylipAcc} =~ /P/i);

				$nameLabel{$phylipAcc} = substr($phylipSeq{$i}, 0, length($phylipSeq{$i}) - $phylipChars);
					$nameLabel{$phylipAcc} =~ s/\s+/_/g;
					$nameLabel{$phylipAcc} =~ s/\_+/_/g;
			}
		}
	}

# Version history
#
#	v1.0 (October 4, 2016)
#		- initial release
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is available at
# http://www.gnu.org/copyleft/gpl.html or by writing to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Fifth Floor, Boston, MA, 02110-1301, USA.
